from sqlalchemy import *
import fixture
from migrate.changeset import *
from migrate.changeset import _ColumnDelta
from sqlalchemy.databases import information_schema

class AddDropColumn(fixture.DB):
    level=fixture.DB.CONNECT
    meta = DynamicMetaData()
    # We'll be adding the 'data' column
    table_name = 'tmp_adddropcol'

    def setUp(self):
        #self.engine.echo=True
        self.table = Table(self.table_name,self.meta,
            Column('id',Integer,primary_key=True),
            redefine=True
        )
        super(AddDropColumn,self).setUp()
        self.meta.connect(self.engine)
        if self.engine.has_table(self.table.name):
            self.table.drop()
        self.table.create()
    def tearDown(self):
        super(AddDropColumn,self).tearDown()
        self.conn.close()
        if self.engine.has_table(self.table.name):
            self.table.drop()
        self.table.deregister()

    def run_(self,create_column_func,drop_column_func,*col_p,**col_k):
        col_name = 'data'
        def _assert_numcols(expected,type_):
            result = len(self.table.c)
            self.assertEquals(result,expected,
                "# %s cols incorrect: %s != %s"%(type_,result,expected))
            if not col_k.get('primary_key',None):
                return
            # new primary key: check its length too
            result = len(self.table.primary_key)
            self.assertEquals(result,expected,
                "# %s pks incorrect: %s != %s"%(type_,result,expected))
        def assert_numcols(expected):
            # number of cols should be correct in table object and in database
            # Changed: create/drop shouldn't mess with the objects
            #_assert_numcols(expected,'object')
            # Detect # database cols via autoload
            self.table = self.refresh_table(self.table)
            self.table.deregister()
            self.table=Table(self.table_name,self.meta,redefine=True,autoload=True)
            _assert_numcols(expected,'database')
        assert_numcols(1)
        if len(col_p) == 0:
            col_p = [String]
        col = Column(col_name,*col_p,**col_k)
        create_column_func(col)
        #create_column(col,self.table)
        assert_numcols(2)
        self.assertEquals(getattr(self.table.c,col_name),col)
        #drop_column(col,self.table)
        drop_column_func(getattr(self.table.c,col_name))
        assert_numcols(1)

    @fixture.supported('postgres')
    def test_undefined(self):
        """Add/drop columns not yet defined in the table"""
        def add_func(col):
            return create_column(col,self.table)
        def drop_func(col):
            return drop_column(col,self.table)
        return self.run_(add_func,drop_func)

    @fixture.supported('postgres')
    def test_defined(self):
        """Add/drop columns already defined in the table"""
        def add_func(col):
            # Pretend the table was declared with this column in it, *before*
            # the alter statement
            self.table.deregister()
            self.table = Table(self.table_name,self.meta,
                Column('id',Integer,primary_key=True),
                col,
                redefine=True
            )
            return create_column(col,self.table)
        def drop_func(col):
            return drop_column(col,self.table)
        return self.run_(add_func,drop_func)

    @fixture.supported('postgres')
    def test_method_bound(self):
        """Add/drop columns via column methods; columns bound to a table
        ie. no table parameter passed to function
        """
        def add_func(col):
            self.table.append_column(col)
            return col.create()
        def drop_func(col):
            self.table.append_column(col)
            return col.drop()
        return self.run_(add_func,drop_func)

    @fixture.supported('postgres')
    def test_method_notbound(self):
        """Add/drop columns via column methods; columns not bound to a table"""
        def add_func(col):
            return col.create(self.table)
        def drop_func(col):
            return col.drop(self.table)
        return self.run_(add_func,drop_func)

    @fixture.supported('postgres')
    def test_tablemethod_obj(self):
        """Add/drop columns via table methods; by column object"""
        def add_func(col):
            return self.table.create_column(col)
        def drop_func(col):
            return self.table.drop_column(col)
        return self.run_(add_func,drop_func)

    @fixture.supported('postgres')
    def test_tablemethod_name(self):
        """Add/drop columns via table methods; by column name"""
        def add_func(col):
            # must be bound to table
            self.table.append_column(col)
            return self.table.create_column(col.name)
        def drop_func(col):
            # Not necessarily bound to table
            return self.table.drop_column(col.name)
        return self.run_(add_func,drop_func)
        
    @fixture.supported('postgres')
    def test_byname(self):
        """Add/drop columns via functions; by table object and column name"""
        def add_func(col):
            self.table.append_column(col)
            return create_column(col.name,self.table)
        def drop_func(col):
            return drop_column(col.name,self.table)
        return self.run_(add_func,drop_func)

    @fixture.supported('postgres')
    def test_fk(self):
        """Can create columns with foreign keys"""
        reftable = Table('tmp_ref',self.meta,
            Column('id',Integer,primary_key=True)
        )
        def add_func(col):
            # create FK's target
            if self.engine.has_table(reftable.name):
                reftable.drop()
            reftable.create()
            self.table.append_column(col)
            return create_column(col.name,self.table)
        def drop_func(col):
            ret = drop_column(col.name,self.table)
            if self.engine.has_table(reftable.name):
                reftable.drop()
            return ret
        return self.run_(add_func,drop_func,Integer,ForeignKey('tmp_ref.id'))

    @fixture.supported('postgres')
    def xtest_pk(self):
        """Can create/drop primary key columns
        Not supported
        """
        def add_func(col):
            create_column(col,self.table)
        def drop_func(col):
            drop_column(col,self.table)
        # Primary key length is checked in run_
        return self.run_(add_func,drop_func,Integer,primary_key=True)

class Rename(fixture.DB):
    level=fixture.DB.CONNECT
    meta = DynamicMetaData()

    def setUp(self):
        self.meta.connect(self.engine)

    @fixture.supported('postgres')
    def test_rename_table(self):
        """Tables can be renamed"""
        #self.engine.echo=True
        name1 = 'name_one'
        name2 = 'name_two'
        xname1 = 'x'+name1
        xname2 = 'x'+name2
        self.column = Column(name1,Integer)
        self.table = Table(name1,self.meta,self.column)
        self.index = Index(xname1,self.column,unique=False)
        if self.engine.has_table(self.table.name):
            self.table.drop()
        if self.engine.has_table(name2):
            tmp = Table(name2,self.meta,autoload=True)
            tmp.drop()
            tmp.deregister()
            del tmp
        self.table.create()

        def assert_table_name(expected,skip_object_check=False):
            if not skip_object_check:
                # Table object check
                self.assertEquals(self.table.name,expected)
                newname = None
            else:
                # we know the object's name isn't consistent: just assign it
                newname = expected
            # Table DB check
            table = self.refresh_table(self.table,newname)
            self.assertEquals(self.table.name,expected)
        def assert_index_name(expected,skip_object_check=False):
            if not skip_object_check:
                # Index object check
                self.assertEquals(self.index.name,expected)
            else:
                # object is inconsistent
                self.index.name = expected
            # Index DB check
            #TODO

        try:
            # Table renames
            assert_table_name(name1)
            rename_table(self.table,name2)
            assert_table_name(name2)
            self.table.rename(name1)
            assert_table_name(name1)
            # ..by just the string
            rename_table(name1,name2,engine=self.engine)
            assert_table_name(name2,True)   # object not updated
    
            # Index renames
            assert_index_name(xname1)
            rename_index(self.index,xname2,engine=self.engine)
            assert_index_name(xname2)
            self.index.rename(xname1)
            assert_index_name(xname1)
            # ..by just the string
            rename_index(xname1,xname2,engine=self.engine)
            assert_index_name(xname2,True)

            # Column renames
            # Nope - we're using alter_column for this
            #assert_column_name(name1)
            #rename_column(self.column,name2)
            #assert_column_name(name2)
            #self.column.rename(name1)
            #assert_column_name(name1)
    
        finally:
            self.index.drop()
            self.table.drop()

class ColumnChange(fixture.DB):
    level=fixture.DB.CONNECT
    meta = DynamicMetaData()
    table_name = 'tmp_colchange'

    def setUp(self):
        fixture.DB.setUp(self)
        self.meta.connect(self.engine)
        self.table = Table(self.table_name,self.meta,
            Column('id',Integer,primary_key=True),
            Column('data',String,PassiveDefault("tluafed"),nullable=True),
            redefine=True
        )
        if self.engine.has_table(self.table.name):
            self.table.drop()
        self.table.create()
    def tearDown(self):
        #self.engine.echo=False
        self.table.drop()
        fixture.DB.tearDown(self)

    @fixture.supported('postgres')
    def test_rename(self):
        """Can rename a column"""
        def num_rows(col,content):
            return len(list(self.table.select(col==content).execute()))
        # Table content should be preserved in changed columns
        content = "fgsfds"
        self.engine.execute(self.table.insert(),data=content)
        self.assertEquals(num_rows(self.table.c.data,content),1)

        # ...as a function, given a column object and the new name
        alter_column(self.table.c.data, name='atad')
        self.table = self.refresh_table(self.table)
        self.assertRaises(AttributeError,getattr,self.table.c,'data')
        self.table.c.atad   # Should not raise exception
        self.assertEquals(num_rows(self.table.c.atad,content),1)

        # ...as a method, given a new name
        self.table.c.atad.alter(name='data')
        self.table = self.refresh_table(self.table)
        self.assertRaises(AttributeError,getattr,self.table.c,'atad')
        self.table.c.data # Should not raise exception
        self.assertEquals(num_rows(self.table.c.data,content),1)

        # ...as a function, given a new object
        col = Column('atad',String)
        alter_column(self.table.c.data, col)
        self.table = self.refresh_table(self.table)
        self.assertRaises(AttributeError,getattr,self.table.c,'data')
        self.table.c.atad   # Should not raise exception
        self.assertEquals(num_rows(self.table.c.atad,content),1)

        # ...as a method, given a new object
        col = Column('data',String)
        self.table.c.atad.alter(col)
        self.table = self.refresh_table(self.table)
        self.assertRaises(AttributeError,getattr,self.table.c,'atad')
        self.table.c.data   # Should not raise exception
        self.assertEquals(num_rows(self.table.c.data,content),1)

        # ...as a function, given an old name, new name, table
        alter_column('data','atad',table=self.table_name,engine=self.engine)
        self.table = self.refresh_table(self.table)
        self.assertRaises(AttributeError,getattr,self.table.c,'data')
        self.table.c.atad   # Should not raise exception
        self.assertEquals(num_rows(self.table.c.atad,content),1)
        
    @fixture.supported('postgres')
    def xtest_fk(self):
        """Can add/drop foreign key constraints to/from a column
        Not supported
        """
        self.assert_(self.table.c.data.foreign_key is None)

        # add
        self.table.c.data.alter(foreign_key=ForeignKey(self.table.c.id))
        self.table = self.refresh_table(self.table)
        self.assert_(self.table.c.data.foreign_key is not None)

        # drop
        self.table.c.data.alter(foreign_key=None)
        self.table = self.refresh_table(self.table)
        self.assert_(self.table.c.data.foreign_key is None)

    @fixture.supported('postgres')
    def test_type(self):
        """Can change a column's type"""
        # Entire column definition given
        self.table.c.data.alter(Column('data',String(42)))
        self.table = self.refresh_table(self.table)
        self.assert_(isinstance(self.table.c.data.type,String))
        self.assertEquals(self.table.c.data.type.length,42)

        # Just the new type
        self.table.c.data.alter(type=String(21))
        self.table = self.refresh_table(self.table)
        self.assert_(isinstance(self.table.c.data.type,String))
        self.assertEquals(self.table.c.data.type.length,21)

        # Different type
        self.assert_(isinstance(self.table.c.id.type,Integer))
        self.table.c.id.alter(type=String)
        self.table = self.refresh_table(self.table)
        self.assert_(isinstance(self.table.c.id.type,String))

    @fixture.supported('postgres')
    def test_default(self):
        """Can change a column's default value (PassiveDefaults only)
        Only PassiveDefaults are changed here: others are managed by the 
        application / by SA
        """
        #self.engine.echo=True
        self.assertEquals(self.table.c.data.default.arg,'tluafed')

        # Just the new default
        default = 'my_default'
        self.table.c.data.alter(default=PassiveDefault(default))
        self.table = self.refresh_table(self.table)
        #self.assertEquals(self.table.c.data.default.arg,default)
        # TextClause returned by autoload
        self.assert_(default in str(self.table.c.data.default.arg))

        # Column object
        default = 'your_default'
        self.table.c.data.alter(Column('data',String,default=PassiveDefault(default)))
        self.table = self.refresh_table(self.table)
        self.assert_(default in str(self.table.c.data.default.arg))

        # Remove default
        self.table.c.data.alter(default=None)
        self.table = self.refresh_table(self.table)
        self.assert_(self.table.c.data.default is None)

    @fixture.supported('postgres')
    def test_null(self):
        """Can change a column's null constraint"""
        self.assertEquals(self.table.c.data.nullable,True)
        
        # Column object
        self.table.c.data.alter(Column('data',String,nullable=False))
        self.table.nullable=None
        self.table = self.refresh_table(self.table)
        self.assertEquals(self.table.c.data.nullable,False)

        # Just the new status
        self.table.c.data.alter(nullable=True)
        self.table = self.refresh_table(self.table)
        self.assertEquals(self.table.c.data.nullable,True)

    @fixture.supported('postgres')
    def xtest_pk(self):
        """Can add/drop a column to/from its table's primary key
        Not supported
        """
        self.assertEquals(len(self.table.primary_key),1)

        # Entire column definition
        self.table.c.data.alter(Column('data',String,primary_key=True))
        self.table = self.refresh_table(self.table)
        self.assertEquals(len(self.table.primary_key),2)

        # Just the new status
        self.table.c.data.alter(primary_key=False)
        self.table = self.refresh_table(self.table)
        self.assertEquals(len(self.table.primary_key),1)

class TestColumnDelta(fixture.Base):
    def test_deltas(self):
        def mkcol(name='id',type=String,*p,**k):
            return Column(name,type,*p,**k)
        col_orig = mkcol(primary_key=True)

        def verify(expected,original,*p,**k):
            delta = _ColumnDelta(original,*p,**k)
            result = delta.keys()
            result.sort()
            self.assertEquals(expected,result)
            return delta

        verify([],col_orig)
        verify(['name'],col_orig,'ids')
        # Parameters are always executed, even if they're 'unchanged'
        # (We can't assume given column is up-to-date)
        verify(['name','primary_key','type'],col_orig,'id',Integer,primary_key=True)
        verify(['name','primary_key','type'],col_orig,name='id',type=Integer,primary_key=True)

        # Can compare two columns and find differences
        col_new = mkcol(name='ids',primary_key=True)
        verify([],col_orig,col_orig)
        verify(['name'],col_orig,col_orig,'ids')
        verify(['name'],col_orig,col_orig,name='ids')
        verify(['name'],col_orig,col_new)
        verify(['name','type'],col_orig,col_new,type=String)
        # Change name, given an up-to-date definition and the current name
        delta = verify(['name'],col_new,current_name='id')
        self.assertEquals(delta.get('name'),'ids')
        # Change other params at the same time
        verify(['name','type'],col_new,current_name='id',type=String)
        # Type comparisons
        verify([],mkcol(type=String),mkcol(type=String))
        verify(['type'],mkcol(type=String),mkcol(type=Integer))
        verify(['type'],mkcol(type=String),mkcol(type=String(42)))
        verify([],mkcol(type=String(42)),mkcol(type=String(42)))
        verify(['type'],mkcol(type=String(24)),mkcol(type=String(42)))
        # Other comparisons
        verify(['primary_key'],mkcol(nullable=False),mkcol(primary_key=True))
        # PK implies nullable=False
        verify(['nullable','primary_key'],mkcol(nullable=True),mkcol(primary_key=True))
        verify([],mkcol(primary_key=True),mkcol(primary_key=True))
        verify(['nullable'],mkcol(nullable=True),mkcol(nullable=False))
        verify([],mkcol(nullable=True),mkcol(nullable=True))
        verify(['default'],mkcol(default=None),mkcol(default='42'))
        verify([],mkcol(default=None),mkcol(default=None))
        verify([],mkcol(default='42'),mkcol(default='42'))

if __name__=='__main__':
    fixture.main()
